/**
 * @file Firmata.c
 * @brief Firmata Protocol
 */

#include <board.h>
#include <Std_Types.h>
#include <BSW/IODrivers/Dio.h>
#include <BSW/IODrivers/Sci.h>

#define __FIRMATA_C
#include "Firmata.h"


static void Fm_Pin_DiWrite(uint8_t ui8t_pin, boolean bot_state);
static boolean Fm_Pin_DiRead(uint8_t ui8t_pin);
static uint32_t Fm_Pin_AnRead(uint8_t ui8t_anPort);

static void Fm_Port_Write(uint8_t ui8t_port, uint8_t ui8t_value);
static uint8_t Fm_Port_Read(uint8_t ui8t_port);

static uint8_t Fm_GetPacketSize(uint8_t rxData);

static void Firmata_VersionSend(void);
static boolean Fm_ReceivePacket(void);
static boolean Fm_SetPacketBuffer(uint8_t rxData);
static void Fm_putchar(uint8_t n);
static boolean Fm_getchar(uint8_t *n);

static void Fm_Execute_Command(void);
static void Fm_Execute_OfflineCommand(void);
static void Fm_Exec_SystemMessage(void);
static void Fm_Exec_Reset(void);
static void Fm_Exec_ProtocolVersion(void);
static void Fm_Exec_SysExMessage(void);

static void Fm_Exec_CapabilityQuery(void);
static void Fm_Exec_AnMapQuery(void);
static void Fm_Exec_PinStateQuery(uint8_t ui8t_pin);
static void Fm_Exec_ExtendedAnWrite(uint8_t ui8t_pin);
static void Fm_Exec_ReportFirmware(void);
static void Fm_Exec_String(void);


static void Do_Report(void);
static void Do_Report_Port(void);
static void Do_Report_Analog(void);
static void Firmata_UpdateAckPort(void);
static void Firmata_CheckCon(void);
void Firmata_TxFlush(void);


/* 受信バッファ */
static uint8_t Fm_RxPacketSize;
static uint8_t Fm_RxCnt;
uint8_t Fm_RxLength;
uint8_t Fm_RxBuffer[BUFFER_SIZE];
static boolean Fm_RxIsValid;

/* 接続状態 */
#define COMTYPE_USB     (0)
#define COMTYPE_SCI1    (1)
static uint8_t Fm_ComType = COMTYPE_USB;
static boolean Firmata_IsConActivePrev = FALSE;
static boolean Fm_IsEnabled = FALSE;
static boolean Fm_IsVersionSent;

/* AckPortタイマ */
/* RXマイコン起動ポートをLo保持する時間 */
#define Fm_AckPortTimer_VALUE       (100 / 10)
static uint16_t Fm_AckPortTimer;


/* 接続待ちタイマ */
/* Win32API CreateFile～PurgeCommで通信バッファ初期化、                     */
/* node-serialportでdataイベントハンドラを登録完了するまでの想定時間を設定 */
/* 上記期間にProtocolVersionを送信した場合、PC側で取りこぼしが発生するため、一定時間待ってからVersionResponseを送信する */
#define Fm_VersionSendTimer_VALUE   (1000 / 10)
static uint16_t Fm_VersionSendTimer;


/* レポート出力周期 */
#define Fm_ReportTimerCycle_Init (2)
static uint8_t Fm_ReportTimerCycle;
static uint8_t Fm_ReportTimerCount;

/* ログ出力 */
boolean Fm_IsLogOutput;


/* Pin設定 */
uint8_t Fm_PinModeIdx[FM_PIN_NUM];
uint8_t Fm_Pin2AnMapping[FM_PIN_NUM];

/* Port設定 */
boolean Fm_PortReportIsEnable[FM_PORT_NUM];

/* Analog設定 */
uint8_t Fm_An2PinMapping[FM_AN_NUM];
boolean Fm_AnReportIsEnable[FM_AN_NUM];


/* 接続ポート状態 */
boolean Fm_DtePortPrev;
boolean Fm_ConnectState;
boolean Fm_ConHistory;

/**
 * @brief 初期化
 */
void Firmata_Init(void)
{
    Fm_ComType = COMTYPE_SCI1;

    if (Fm_ComType == COMTYPE_USB)
    {
        /*USBDrv_Init();*/
        
        /*USBDrv_SetDeviceIdx(IDX_1);*/
        /*USBDrv_Attach();*/
    }
    
    Fm_ConnectState = FALSE;
    Fm_ConHistory = FALSE;
    Fm_DtePortPrev = FALSE;

    Fm_IsEnabled = TRUE;
    Fm_AckPortTimer = 0;
    Fm_IsVersionSent = FALSE;
    Fm_VersionSendTimer = 0;
    Fm_ReportTimerCycle = Fm_ReportTimerCycle_Init;
    Fm_ReportTimerCount = 0;
    Fm_Exec_Reset();
    
    Fm_OnInit();
}


/**
 * @brief 解放
 */
void Firmata_DeInit(void)
{
    if (Fm_ComType == COMTYPE_USB)
    {
        /*USBDrv_Detach();*/
    }
}


/**
 * @brief Pin Digital Write
 */
static void Fm_Pin_DiWrite(uint8_t ui8t_pin, boolean bot_state)
{
    if (FM_PIN_NUM <= ui8t_pin) {
        return;
    }
    
    if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode == FM_PINMODE_OUTPUT) {
        if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp != NULL) {
            if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_OUTPUT_VI)
            {
                ((void (*)(int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(bot_state);
            }
            else if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_OUTPUT_II)
            {
                ((void (*)(int, int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(
                    Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].arg1,
                    bot_state
                );
            }
            else
            {
            }
        }
    }
}


/**
 * @brief Pin Digital Read
 */
static boolean Fm_Pin_DiRead(uint8_t ui8t_pin)
{
    if (FM_PIN_NUM <= ui8t_pin) {
        return (FALSE);
    }
    
    if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode == FM_PINMODE_INPUT) {
        if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp != NULL) {
            if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_INPUT_IV)
            {
                return ((int (*)(void))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))();
            }
            else if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_INPUT_II)
            {
                return ((int (*)(int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].arg1);
            }
            else
            {
            }
        }
    }
    return (FALSE);
}


/**
 * @brief Pin Analog Read
 */
static uint32_t Fm_Pin_AnRead(uint8_t ui8t_anPort)
{
    uint8_t ui8t_pin;
    
    if (FM_AN_NUM <= ui8t_anPort) {
        return (VAL_0);
    }
    
    ui8t_pin = Fm_An2PinMapping[ui8t_anPort];
    
    if ((ui8t_pin < FM_PORT_NUM) && (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode == FM_PINMODE_ANALOG)) {
        if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp != NULL)
        {
            if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_INPUT_IV)
            {
                return ((int (*)(void))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))();
            }
            else if (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_INPUT_II)
            {
                return ((int (*)(int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].arg1);
            }
            else
            {
            }
        }
    }
    return (VAL_0);
}


/**
 * @brief Pin Analog Write
 */
static void Fm_Pin_AnWrite(uint8_t ui8t_pin, uint16_t ui16t_value)
{
    uint8_t ui8t_mode;

    if (FM_PIN_NUM <= ui8t_pin) {
        return;
    }

    ui8t_mode = Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode;

    if (ui8t_mode == FM_PINMODE_PWM) {
        if ((Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_OUTPUT_II) &&
            (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp != NULL)) {
            ((void (*)(uint32_t, int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].arg1, ui16t_value);
        }

    } else {
        /* nothing to do. */
    }
}


/**
 * @brief Port Write
 */
static void Fm_Port_Write(uint8_t ui8t_port, uint8_t ui8t_value)
{
    uint8_t ui8t_bit;
    uint8_t ui8t_pin;
    
    for (ui8t_bit = IDX_0; ui8t_bit < IDX_8; ui8t_bit++) {
        ui8t_pin = (ui8t_port * IDX_8) + ui8t_bit;

        Fm_Pin_DiWrite(ui8t_pin, ui8t_value & BIT_0);
        ui8t_value >>= IDX_1;
    }
}


/**
 * @brief Port Read
 */
static uint8_t Fm_Port_Read(uint8_t ui8t_port)
{
    uint8_t ui8t_value;
    uint8_t ui8t_pin;
    uint8_t ui8t_bit;
    
    ui8t_value = VAL_0;
    for (ui8t_bit = IDX_0; ui8t_bit < IDX_8; ui8t_bit++) {
        ui8t_pin = (ui8t_port * IDX_8) + ui8t_bit;
        
        ui8t_value >>= IDX_1;
        ui8t_value |= Fm_Pin_DiRead(ui8t_pin) << IDX_7;
    }
    
    return (ui8t_value);
}


/**
 * @brief GetPacketSize
 */
static uint8_t Fm_GetPacketSize(uint8_t rxData)
{
    uint8_t cmd;

    
    cmd = rxData & FM_MSGTYPE_MASK;

    switch (cmd) {
        case FM_MSGTYPE_REPORT_AN_PIN:
        case FM_MSGTYPE_REPORT_DI_PIN:
            return (IDX_2);
        case FM_MSGTYPE_AN_IO:
        case FM_MSGTYPE_DI_IO:
            return (IDX_3);
        case FM_MSGTYPE_SYSTEM:                 /* 0xF* */
            switch (rxData) {
                case FM_MSGTYPE_SYSEX_START:
                case FM_MSGTYPE_SYSEX_END:
                    return (IDX_0);
                case FM_MSGTYPE_SYSTEM_RESET:
                case FM_MSGTYPE_PROTOCOL_VERSION:
                    return (IDX_1);
                case FM_MSGTYPE_SET_PIN_MODE:
                    return (IDX_3);
                default:
                    break;
            }
            break;
        default:
            break;
    }
    
    return (IDX_0);
}


/**
 * @brief Send Version
 */
static void Firmata_VersionSend(void)
{
    Fm_Exec_ProtocolVersion();
    Fm_Exec_ReportFirmware();
}


/**
 * @brief Receive Packet
 */
static boolean Fm_ReceivePacket(void)
{
    uint8_t rxByte;
    
    while (Fm_getchar(&rxByte) == TRUE) {
        /* パケットバッファにデータをセットし、1パケット構築完了 */
        if (Fm_SetPacketBuffer(rxByte) == TRUE) {
            return (TRUE);
        }
    }
    
    /* 受信データがなくなった */
    return (FALSE);
}


/**
 * @brief Set Packet Buffer
 */
static boolean Fm_SetPacketBuffer(uint8_t rxData)
{
    /* ステータスビットが立っている */
    if ((rxData & BIT_7) != VAL_0) {
        /* SysExコマンドを受信中でない かつ 受信したコマンドがSYSEX_ENDではない */
        if ((Fm_RxPacketSize == IDX_0) && (rxData != FM_MSGTYPE_SYSEX_END)) {
            /* データを最初から受信しなおす */
            Fm_RxCnt = IDX_0;
            Fm_RxIsValid = TRUE;
        }
    }
    
    /* 1Byte目の受信時 */
    if (Fm_RxCnt == IDX_0) {
        /* パケットサイズを取得 */
        Fm_RxPacketSize = Fm_GetPacketSize(rxData);
    }

    /* バッファに収まるなら */
    if (Fm_RxCnt < BUFFER_SIZE) {
        /* バッファに格納 */
        Fm_RxBuffer[Fm_RxCnt++] = rxData;
        
    } else {
        Fm_RxIsValid = FALSE;
    }
    
    
    
    /* Finish Rx */
    /* 可変長メッセージかつSysEx Endではない */
    /* または固定長メッセージで長さ分受信 */
    if (((Fm_RxPacketSize == IDX_0) && (rxData == FM_MSGTYPE_SYSEX_END)) ||
        ((Fm_RxPacketSize != IDX_0) && (Fm_RxPacketSize <= Fm_RxCnt))) {
        /* 受信データ長を設定 */
        Fm_RxLength = Fm_RxCnt;
        
        /* 次の受信インデックスを初期化 */
        Fm_RxCnt = IDX_0;
        
        /* 受信データ有効なら */
        if (Fm_RxIsValid == TRUE) {
            /* 戻り値 = TRUE */
            return (TRUE);
        }
    }
    
    /* 戻り値 = FALSE */
    return (FALSE);
}


/**
 * @brief putchar
 */
static void Fm_putchar(uint8_t n)
{
    if (Fm_ComType== COMTYPE_USB)
    {
        /*USB_CDC_putchar(n);*/
        
#if FIRMATA_DEBUG_COM
        if (Fm_IsLogOutput == TRUE) {
            Com_SendHex(n, 2);
        }
#endif
    }
    else if (Fm_ComType== COMTYPE_SCI1)
    {
        Sci_Send1BAsync(n);
    }
    else
    {
    }
}


/**
 * @brief gutchar
 */
static boolean Fm_getchar(uint8_t *n)
{
    if (Fm_ComType == COMTYPE_USB)
    {
        /*return (USB_CDC_getchar(n));*/
    }
    else if (Fm_ComType == COMTYPE_SCI1)
    {
        return (Sci_Recv1BAsync(n));
    }
    else
    {
    }
    
    return (FALSE);
}


/**
 * @brief Execute Command
 */
static void Fm_Execute_Command(void)
{
    uint8_t cmd;
    uint8_t ch;
    uint32_t i;
    
    cmd = Fm_RxBuffer[IDX_0] & FM_MSGTYPE_MASK;
    ch = Fm_RxBuffer[IDX_0] & FM_CHANNEL_MASK;
    
#if FIRMATA_DEBUG_COM
    Com_SendStr("\r\nCmd=");
#endif
    
    for (i = 0; i < Fm_RxLength; i++) {
#if FIRMATA_DEBUG_COM
        Com_SendHex(Fm_RxBuffer[i], 2);
#endif
    }
#if FIRMATA_DEBUG_COM
    Com_SendStr(", Res=");
#endif

    switch (cmd) {
    case FM_MSGTYPE_SYSTEM:
        Fm_Exec_SystemMessage();
        break;
    case FM_MSGTYPE_DI_IO:
        Fm_Port_Write(ch, (Fm_RxBuffer[IDX_1] | (Fm_RxBuffer[IDX_2] << IDX_7)));
        break;
    case FM_MSGTYPE_REPORT_AN_PIN:
        Fm_AnReportIsEnable[ch] = Fm_RxBuffer[IDX_1] & TRUE;
        break;
    case FM_MSGTYPE_REPORT_DI_PIN:
        if (Fm_PortIsReadable[ch] == TRUE) {
            Fm_PortReportIsEnable[ch] = Fm_RxBuffer[IDX_1] & TRUE;
        }
        break;
    case FM_MSGTYPE_AN_IO:
        Fm_Pin_AnWrite(ch, (Fm_RxBuffer[IDX_1] | (Fm_RxBuffer[IDX_2] << IDX_7)));
        break;
    }
}


/**
 * @brief Execute System Message
 */
static void Fm_Exec_SystemMessage(void)
{
    switch (Fm_RxBuffer[IDX_0]) {
    case FM_MSGTYPE_SYSTEM_RESET:
        Fm_Exec_Reset();
        break;
    case FM_MSGTYPE_SYSEX_START:
        Fm_Exec_SysExMessage();
        break;
    case FM_MSGTYPE_PROTOCOL_VERSION:
        Fm_Exec_ProtocolVersion();
        break;
    default:
        break;
    }
}


/**
 * @brief Reset Command
 */
static void Fm_Exec_Reset(void)
{
    uint8_t ui8t_pin;
    uint8_t ui8t_port;
    uint8_t ui8t_modeIdx;
    uint8_t ui8t_anMapIdx;
    
    Fm_RxPacketSize = IDX_0;
    Fm_RxCnt = IDX_0;
    Fm_RxIsValid = FALSE;
    
    
    /* Analog PortのMappingを初期化 */
    for (ui8t_anMapIdx = IDX_0; ui8t_anMapIdx < FM_AN_NUM; ui8t_anMapIdx++) {
        /* Pin Mappingなし、Report無しに設定 */
        Fm_An2PinMapping[ui8t_anMapIdx] = FM_PIN_UNASSIGNED;
        Fm_AnReportIsEnable[ui8t_anMapIdx] = FALSE;
    }
    
    /* PortのReport設定を初期化 */
    for (ui8t_port = IDX_0; ui8t_port < FM_PORT_NUM; ui8t_port++) {
        Fm_PortReportIsEnable[ui8t_port] = FALSE;
        
    }
    
    ui8t_anMapIdx = IDX_0;
    for (ui8t_pin = IDX_0; ui8t_pin < FM_PIN_NUM; ui8t_pin++) {
        Fm_PinModeIdx[ui8t_pin] = IDX_0;
        Fm_Pin2AnMapping[ui8t_pin] = FM_PIN_UNASSIGNED;
        
        ui8t_modeIdx = IDX_0;
        while (Fm_Pins[ui8t_pin][ui8t_modeIdx].mode != FM_PINMODE_IGNORE) {
            if ((ui8t_anMapIdx < FM_AN_NUM) && (Fm_Pins[ui8t_pin][ui8t_modeIdx].mode == FM_PINMODE_ANALOG)) {
                Fm_An2PinMapping[ui8t_anMapIdx] = ui8t_pin;
                Fm_Pin2AnMapping[ui8t_pin] = ui8t_anMapIdx;
                ui8t_anMapIdx++;
                break;
            }
            
            ui8t_modeIdx++;
        }
    }
    
    Fm_OnReset();
}


/**
 * @brief ProtocolVersion
 */
static void Fm_Exec_ProtocolVersion(void)
{
    Fm_putchar(FM_MSGTYPE_PROTOCOL_VERSION);
    Fm_putchar(FM_PROTOCOL_VERSION_H);
    Fm_putchar(FM_PROTOCOL_VERSION_L);
}


/**
 * @brief SysExMessage
 */
static void Fm_Exec_SysExMessage(void)
{
    switch (Fm_RxBuffer[IDX_1]) {
    case FM_SYSEXCMD_AN_MAP_QUERY:
        Fm_Exec_AnMapQuery();
        break;
    case FM_SYSEXCMD_CAPABILITY_QUERY:
        Fm_Exec_CapabilityQuery();
        break;
    case FM_SYSEXCMD_PIN_STATE_QUERY:
        /* Check Format */
        if (Fm_RxLength == IDX_4) {
            Fm_Exec_PinStateQuery(Fm_RxBuffer[IDX_2]);
        }
        break;
    case FM_SYSEXCMD_EXTENDED_AN:
        Fm_Exec_ExtendedAnWrite(Fm_RxBuffer[IDX_2]);
        break;
    case FM_SYSEXCMD_REPORT_FIRMWARE:
        if (Fm_RxLength == IDX_3) {
            Fm_Exec_ReportFirmware();
        }
        break;
    case FM_SYSEXCMD_STRING_DATA:
        Fm_Exec_String();
        break;
    default:
        break;
        
    }
}


/**
 * @brief Analog Mapping Query
 */
static void Fm_Exec_AnMapQuery(void)
{
    uint8_t ui8t_pin;

    Fm_putchar(FM_MSGTYPE_SYSEX_START);
    Fm_putchar(FM_SYSEXCMD_AN_MAP_RESPONSE);
    
    for (ui8t_pin = IDX_0; ui8t_pin < FM_PIN_NUM; ui8t_pin++)
    {
        Fm_putchar(Fm_Pin2AnMapping[ui8t_pin]);
    }
    
    Fm_putchar(FM_MSGTYPE_SYSEX_END);
}


/**
 * @brief CapabilityQuery
 */
static void Fm_Exec_CapabilityQuery(void)
{
    uint8_t ui8t_pin;
    uint8_t ui8t_modeIdx;
    
    Fm_putchar(FM_MSGTYPE_SYSEX_START);
    Fm_putchar(FM_SYSEXCMD_CAPABILITY_RESPONSE);

    for (ui8t_pin = IDX_0; ui8t_pin < FM_PIN_NUM; ui8t_pin++) {
        ui8t_modeIdx = Fm_PinModeIdx[ui8t_pin];
        
        while (Fm_Pins[ui8t_pin][ui8t_modeIdx].mode != FM_PINMODE_IGNORE) {
            Fm_putchar(Fm_Pins[ui8t_pin][ui8t_modeIdx].mode);
            Fm_putchar(Fm_Pins[ui8t_pin][ui8t_modeIdx].resolution);
            ui8t_modeIdx++;
        }
        
        Fm_putchar(FM_PINMODE_IGNORE);
    }

    Fm_putchar(FM_MSGTYPE_SYSEX_END);
}


/**
 * @brief Pin State Query
 */
static void Fm_Exec_PinStateQuery(uint8_t ui8t_pin)
{
    uint8_t ui8t_mode;
    uint32_t ui32t_value;

    if (FM_PIN_NUM <= ui8t_pin) {
        return;
    }
    
    Fm_putchar(FM_MSGTYPE_SYSEX_START);
    Fm_putchar(FM_SYSEXCMD_PIN_STATE_RESPONSE);
    Fm_putchar(ui8t_pin);
    
    ui8t_mode = Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode;
    Fm_putchar(ui8t_mode);

    if (ui8t_mode == FM_PINMODE_INPUT) {
        ui32t_value = Fm_Pin_DiRead(ui8t_pin);
        Fm_putchar(ui32t_value & 0xFF);

    } else if (ui8t_mode == FM_PINMODE_ANALOG) {
        ui32t_value = Fm_Pin_AnRead(ui8t_pin);
        Fm_putchar( ui32t_value           & 0x7F);
        Fm_putchar((ui32t_value >> IDX_7) & 0x7F);

    } else {
        Fm_putchar(0x00);
    }

    Fm_putchar(FM_MSGTYPE_SYSEX_END);
}


/**
 * @brief ReportFirmware
 */
static void Fm_Exec_ExtendedAnWrite(uint8_t ui8t_pin)
{
    uint8_t ui8t_mode;
    int value;
    
    if (FM_PIN_NUM <= ui8t_pin) {
        return;
    }
    
    ui8t_mode = Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode;

    if (ui8t_mode == FM_PINMODE_PWM) {
        if ((Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].funcType == FM_FUNCTYPE_OUTPUT_II) &&
            (Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp != NULL)) {
            
            
            if (Fm_RxLength == IDX_5) {
                value = (int)Fm_RxBuffer[IDX_3];
            } else if (Fm_RxLength == IDX_6) {
                value = (((int)Fm_RxBuffer[IDX_4]) << 7) | ((int)Fm_RxBuffer[IDX_3]);
            } else {
                return;
            }

            ((void (*)(uint32_t, int))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].fp))(Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].arg1, value);
        }
        
    } else {
        /* nothing to do. */
    }
}


/**
 * @brief ReportFirmware
 */
static void Fm_Exec_ReportFirmware(void)
{
    uint8_t ui8t_idx;

    Fm_putchar(FM_MSGTYPE_SYSEX_START);
    Fm_putchar(FM_SYSEXCMD_REPORT_FIRMWARE);
    for (ui8t_idx = 0; ui8t_idx < sizeof(Fm_FirmwareVersion); ui8t_idx++)
    {
        Fm_putchar(Fm_FirmwareVersion[ui8t_idx]);
    }
    Fm_putchar(FM_MSGTYPE_SYSEX_END);
}


/**
 * @brief String Data
 */
static void Fm_Exec_String(void)
{
    uint16_t ui16t_idx;
    uint16_t ui16t_len;
    char ch;
    char lcdData[LcdDrv_Y_MAX * LcdDrv_X_MAX + 1];
    
    /* len1 = SYSEX STRING_DATA string(per 2bytes) \0 \0 SYSEXEND */
    ui16t_len = ((Fm_RxLength - 3) / 2) - 1;
    if ((Fm_RxBuffer[2 + ui16t_len * 2] != 0x00) || (Fm_RxBuffer[2 + ui16t_len * 2] != 0x00))
    {
        /* string data is invalid */
        return;
    }
    

    /* Special String Data ? */
    if (ui16t_len != 0 &&
        Fm_RxBuffer[2] == 0x00 &&
        Fm_RxBuffer[3] == 0x00)
    {
        /* nothing to do */
        return;
    }

    /* clear lcd buffer */
    for (ui16t_idx = 0; ui16t_idx < (LcdDrv_Y_MAX * LcdDrv_X_MAX) ; ui16t_idx++)
    {
        lcdData[ui16t_idx] = ' ';
    }
    
    /* set lcd data */
    for (ui16t_idx = 0; ui16t_idx < ui16t_len; ui16t_idx++)
    {
        ch = Fm_RxBuffer[2 + ui16t_idx * 2 + 1] << 7 |
                Fm_RxBuffer[2 + ui16t_idx * 2];

        lcdData[ui16t_idx] = ch;
    }
    
    lcdData[ui16t_idx] = '\0';
    
    FmLCD_SetUserText(lcdData);
}


/**
 * @brief Do Report
 */
static void Do_Report(void)
{
    Do_Report_Port();
    Do_Report_Analog();
}


/**
 * @brief Do Port Report
 */
static void Do_Report_Port(void)
{
    uint8_t ui8t_value;
    uint8_t ui8t_port;
    
    for (ui8t_port = IDX_0; ui8t_port <  FM_PORT_NUM; ui8t_port++) {
        if ((Fm_PortIsReadable[ui8t_port] == TRUE) &&
            (Fm_PortReportIsEnable[ui8t_port] == TRUE)) {
            
            Fm_putchar(FM_MSGTYPE_DI_IO | ui8t_port);
            
            ui8t_value = Fm_Port_Read(ui8t_port);
            Fm_putchar(ui8t_value & 0x7F);
            Fm_putchar((ui8t_value >> IDX_7) & 0x7F);
        }
    }
}


/**
 * @brief Do Analog Report
 */
static void Do_Report_Analog(void)
{
    uint32_t ui32t_value;
    uint8_t ui8t_pin;
    
    for (ui8t_pin = IDX_0; ui8t_pin <  FM_AN_NUM; ui8t_pin++) {
        if ((Fm_Pins[ui8t_pin][Fm_PinModeIdx[ui8t_pin]].mode == FM_PINMODE_ANALOG) &&
            (Fm_AnReportIsEnable[ui8t_pin] == TRUE)) {
            
            Fm_putchar(FM_MSGTYPE_AN_IO | ui8t_pin);
            
            ui32t_value = Fm_Pin_AnRead(ui8t_pin);
            Fm_putchar(ui32t_value & 0x7F);
            Fm_putchar((ui32t_value >> IDX_7) & 0x7F);
        }
    }
}


/**
 * @brief Update AckPort
 */
static void Firmata_UpdateAckPort(void)
{
    if (Fm_AckPortTimer < Fm_AckPortTimer_VALUE)
    {
        Fm_AckPortTimer++;
        Dio_WriteChannel(P25, TRUE);
    }
    else
    {
        Dio_WriteChannel(P25, FALSE);
    }
}


/**
 * @brief Check Connect State
 */
static void Firmata_CheckCon(void)
{
    boolean bot_dtePort;

    /* TOOD: ポート要変更 */
    bot_dtePort = Dio_ReadChannel(P31);

    /* 接続中? */
    if (bot_dtePort == FALSE)
    {
        /* DTE Port Hi->Loエッジで接続OK判定 */
        if (Fm_DtePortPrev == TRUE)
        {
            Fm_ConnectState = TRUE;
        }
    }
    /* 切断? */
    else
    {
        Fm_ConnectState = FALSE;
    }

    Fm_DtePortPrev = bot_dtePort;
}


/**
 * @bridef Is Connection Active
 * @return TRUE: active, FALSE: inactive
 */
boolean Firmata_IsConActive(void)
{
    if (Fm_IsEnabled == FALSE)
    {
        return (FALSE);
    }
    
    if (Fm_ComType == COMTYPE_USB)
    {
        /*return (USB_CDC_IsConnected() && USB_CDC_GetDTE());*/
    }
    else if (Fm_ComType == COMTYPE_SCI1)
    {
        return (Fm_ConnectState);
    }
    else
    {
    }
    
    return (FALSE);
}


/**
 * @bridef Is Connection History
 * @return TRUE: connected history exists, FALSE: connected history none
 */
boolean Firmata_IsConHistory(void)
{
    return (Fm_ConHistory);
}


/**
 * @brief TxBuffer Flush
 * 
 */
void Firmata_TxFlush(void)
{
    if (Fm_ComType == COMTYPE_USB)
    {
        /*USB_CDC_txflush();*/
    }
    else
    {
    }
}


/**
 * @brief 10ms task
 */
void Firmata_10msTask(void)
{
    Firmata_UpdateAckPort();
    Firmata_CheckCon();


    /* 接続済み? */
    if (Firmata_IsConActive() == TRUE) {
        /* 接続履歴 = あり */
        Fm_ConHistory = TRUE;

        /* 接続時処理 */
        if (Firmata_IsConActivePrev == FALSE)
        {
            Firmata_IsConActivePrev = TRUE;

            /* 初期化 */
            Fm_Exec_Reset();
            Fm_OnConnect();
        }
        
        /* Version送信 */
        if (Fm_IsVersionSent == FALSE) {
            Fm_VersionSendTimer = (Fm_VersionSendTimer < Fm_VersionSendTimer_VALUE) ? (Fm_VersionSendTimer + 1) : Fm_VersionSendTimer_VALUE;
            
            if (Fm_VersionSendTimer >= Fm_VersionSendTimer_VALUE)
            {
                Firmata_VersionSend();
                Fm_IsVersionSent = TRUE;
            }
        }
        
        /* コマンドを実行 */
        Fm_IsLogOutput = TRUE;
        while (Fm_ReceivePacket() == TRUE) {
            Fm_Execute_Command();
        }
        Fm_IsLogOutput = FALSE;
        
        /* 一定周期毎にレポート出力 */
        Fm_ReportTimerCount++;
        if (Fm_ReportTimerCount >= Fm_ReportTimerCycle)
        {
            Fm_ReportTimerCount = 0;
            
            /* レポート出力を実行 */
            Do_Report();

            
        }
        
        /* 送信バッファをフラッシュ */
        Firmata_TxFlush();
        
    /* 未接続? */
    } else {
        /* 切断時処理 */
        if (Firmata_IsConActivePrev == TRUE)
        {
            Firmata_IsConActivePrev = FALSE;
            
            /* 初期化 */
            Fm_Exec_Reset();
            
            /* Version送信フラグをクリア */
            Fm_IsVersionSent = FALSE;
            
            /* バージョン送信タイマをクリア */
            Fm_VersionSendTimer = 0;
            
            /* 切断処理 */
            Fm_OnDisconnect();
        }

        /* 未接続中の特殊コマンドを実行 */
        Fm_IsLogOutput = TRUE;
        while (Fm_ReceivePacket() == TRUE) {
            Fm_Execute_OfflineCommand();
        }
        Fm_IsLogOutput = FALSE;
    }

    Fm_OnPeriodic();
}

