/**
 * @file Firmata_Cfg.h
 * @brief Firmataプロトコルの各種設定・定数定義
 */
#ifndef __Firmata_Cfg_H
#define __Firmata_Cfg_H


/*------------------------------------------------------------------------------------------------*/
/* 共通定数                                                                                       */
/*------------------------------------------------------------------------------------------------*/
#define FM_MSGTYPE_MASK                 (0xF0u)
#define FM_CHANNEL_MASK                 (0x0Fu)


#define FM_MSGTYPE_AN_IO                (0xE0u)
#define FM_MSGTYPE_DI_IO                (0x90u)
#define FM_MSGTYPE_REPORT_AN_PIN        (0xC0u)
#define FM_MSGTYPE_REPORT_DI_PIN        (0xD0u)

#define FM_MSGTYPE_SYSTEM               (0xF0u)
#define FM_MSGTYPE_SYSEX_START          (0xF0u)
#define FM_MSGTYPE_SET_PIN_MODE         (0xF4u)
#define FM_MSGTYPE_SYSEX_END            (0xF7u)
#define FM_MSGTYPE_PROTOCOL_VERSION     (0xF9u)
#define FM_MSGTYPE_SYSTEM_RESET         (0xFFu)


#define FM_SYSEXCMD_RESERVED            (0x00u)
#define FM_SYSEXCMD_AN_MAP_QUERY        (0x69u)
#define FM_SYSEXCMD_AN_MAP_RESPONSE     (0x6Au)
#define FM_SYSEXCMD_CAPABILITY_QUERY    (0x6Bu)
#define FM_SYSEXCMD_CAPABILITY_RESPONSE (0x6Cu)
#define FM_SYSEXCMD_PIN_STATE_QUERY     (0x6Du)
#define FM_SYSEXCMD_PIN_STATE_RESPONSE  (0x6Eu)
#define FM_SYSEXCMD_EXTENDED_AN         (0x6Fu)
#define FM_SYSEXCMD_SERVO_CONFIG        (0x70u)
#define FM_SYSEXCMD_STRING_DATA         (0x71u)
#define FM_SYSEXCMD_SHIFT_DATA          (0x75u)
#define FM_SYSEXCMD_I2C_REQUEST         (0x76u)
#define FM_SYSEXCMD_I2C_REPLY           (0x77u)
#define FM_SYSEXCMD_I2C_CONFIG          (0x78u)
#define FM_SYSEXCMD_REPORT_FIRMWARE     (0x79u)
#define FM_SYSEXCMD_SAMPLING_INTERVAL   (0x7Au)
#define FM_SYSEXCMD_SYSEX_NON_REALTIME  (0x7Eu)
#define FM_SYSEXCMD_SYSEX_REALTIME      (0x7Fu)


/* Pin Modes */
#define FM_PINMODE_INPUT                (0x00u)
#define FM_PINMODE_OUTPUT               (0x01u)
#define FM_PINMODE_ANALOG               (0x02u)
#define FM_PINMODE_PWM                  (0x03u)
#define FM_PINMODE_SERVO                (0x04u)
#define FM_PINMODE_SHIFT                (0x05u)
#define FM_PINMODE_I2C                  (0x06u)
#define FM_PINMODE_ONEWIRE              (0x07u)
#define FM_PINMODE_STEPPER              (0x08u)
#define FM_PINMODE_SERIAL               (0x0Au)
#define FM_PINMODE_PULLUP               (0x0Bu)
#define FM_PINMODE_PING_READ            (0x75u)
#define FM_PINMODE_UNKNOWN              (0x10u)
#define FM_PINMODE_IGNORE               (0x7Fu)

#define FM_PIN_UNASSIGNED               (0x7Fu)



/*------------------------------------------------------------------------------------------------*/
/* 各種定数定義                                                                                   */
/*------------------------------------------------------------------------------------------------*/
#define FIRMATA_DEBUG_COM       (0)

#define FM_PROTOCOL_VERSION_H   (0x02u)
#define FM_PROTOCOL_VERSION_L   (0x05u)
#define FM_VERSION              " 1.0"


#define MAX_DATA_BYTES          (66u)                       /**< Max SysEx command data length */
#define BUFFER_SIZE             (MAX_DATA_BYTES + 3u)       /**< START_SYSEX(1) + sysex command(1) + 0..MAX_DATA_BYTES + END_SYSEX(1) */

#define FM_PORT_NUM             (16u)                       /**< Num of Port */
#define FM_AN_NUM               (16u)                       /**< Num of Analog Port */
#define FM_PIN_NUM              (40u)                       /**< Num of Pin */

#define FM_FUNCTYPE_UNDEF       (0x00u)                 /**< Function type is UNDEF */
#define FM_FUNCTYPE_INPUT_IV    (0x01u)                 /**< Function type is INPUT int (void) */
#define FM_FUNCTYPE_OUTPUT_VI   (0x02u)                 /**< Function type is OUTPUT void (int value) */
#define FM_FUNCTYPE_INPUT_II    (0x11u)                 /**< Function type is INPUT int (int) */
#define FM_FUNCTYPE_OUTPUT_II   (0x12u)                 /**< Function type is OUTPUT with arg void (uint32_t arg, int value) */


typedef struct {
    uint8_t mode;                                       /**< Port Mode */
    uint8_t resolution;                                 /**< Port Resolution (bit length) */
                                                        /**@note node.js library firmata-io does not support different resolutions for each pins. */
                                                        /**@note The resolution must be set to the same value in the same mode(adc, pwm, etc...). */
    uint8_t funcType;                                   /**< Function Type */
    void *fp;                                           /**< Function Ptr */
    uint32_t arg1;                                      /**< Function Arg */
} Fm_Pin;


/* BT MAC Address */
extern boolean BTMacAddr_IsReceived;
extern uint8_t BTMacAddr[6];


/* BT SPP Numeric */
extern boolean BTSPPNumeric_IsReceived;
extern uint8_t BTSPPNumeric[6];


/*------------------------------------------------------------------------------------------------*/
/* Firmata C                                                                                      */
/*------------------------------------------------------------------------------------------------*/
#ifdef __FIRMATA_C


/* Firmware Version */
static const uint8_t Fm_FirmwareVersion[] = {
        0x01,0x00,
        0x46,0x00,
        0x69,0x00,
        0x72,0x00,
        0x6D,0x00,
        0x61,0x00,
        0x74,0x00,
        0x61,0x00
};


/*------------------------------------------------------------------------------------------------*/
/* Pin設定                                                                                        */
/*------------------------------------------------------------------------------------------------*/
#include "FmLineTrace.h"
#include "FmMotor.h"
#include "FmBuzzer.h"
#include "FmLCD.h"
#include "FmTestApl.h"


/* Fm Definition */
/*
Fm_Pin Fm_PinN[] = {
    {
        FM_PINMODE_ANALOG,       <- Pin Mode
        0x01,
        FM_FUNCTYPE_INPUT_IV,
        &analogRead,
        0x00
    },
    {                            <- PINMODE_IGNORE must locate final!!!
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
*/
/* FM_PINMODE_IGNORE must locate end of PinModes */
/* I/O変更時、Fm_Exec_Resetの処理を見直すこと */
const Fm_Pin Fm_Pin0[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_VOLUME
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin1[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_DISTANCE
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin2[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_BRIGHTNESS
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin3[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_LINE_L
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin4[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_LINE_R
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin5[] = {
    {
        FM_PINMODE_ANALOG,
        14,
        FM_FUNCTYPE_INPUT_II,
        &analogRead,
        PIN_TOGGLE
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin6[] = {
    {
        FM_PINMODE_INPUT,
        0x01,
        FM_FUNCTYPE_INPUT_II,
        &digitalRead,
        PIN_BTN1
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin7[] = {
    {
        FM_PINMODE_INPUT,
        0x01,
        FM_FUNCTYPE_INPUT_II,
        &digitalRead,
        PIN_BTN2
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin8[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED1
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin9[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED2
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin10[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED3
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin11[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED4
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin12[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED5
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin13[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_II,
        &digitalWrite,
        PIN_LED6
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin14[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmMotor_SetSpeed,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin15[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmMotor_SetSpeed,
        0x01
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin16[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmBuzzer_SetFreq,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin17[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &analogWrite,
        PIN_FLED_R
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin18[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &analogWrite,
        PIN_FLED_G
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin19[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &analogWrite,
        PIN_FLED_B
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_PinUndef[] = {
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin32[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_VI,
        &FmLineTrace_SetEnable,
        0
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin33[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmMotor_SetTimer,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin34[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmMotor_SetTimer,
        0x01
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin35[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmLineTrace_SetSpeed,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin36[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmLineTrace_SetThresh,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin37[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmLineTrace_SetValue1,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin38[] = {
    {
        FM_PINMODE_PWM,
        14,
        FM_FUNCTYPE_OUTPUT_II,
        &FmBuzzer_SetTimer,
        0x00
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};
const Fm_Pin Fm_Pin39[] = {
    {
        FM_PINMODE_OUTPUT,
        0x01,
        FM_FUNCTYPE_OUTPUT_VI,
        &FmLCD_SetMonitorEnable,
        0
    },
    {
        FM_PINMODE_IGNORE,
        0x00,
        FM_FUNCTYPE_UNDEF,
        NULL,
        0x00
    }
};


const Fm_Pin *const Fm_Pins[FM_PIN_NUM] = {
    Fm_Pin0,
    Fm_Pin1,
    Fm_Pin2,
    Fm_Pin3,
    Fm_Pin4,
    Fm_Pin5,
    Fm_Pin6,
    Fm_Pin7,
    Fm_Pin8,
    Fm_Pin9,
    Fm_Pin10,
    Fm_Pin11,
    Fm_Pin12,
    Fm_Pin13,
    Fm_Pin14,
    Fm_Pin15,
    Fm_Pin16,
    Fm_Pin17,
    Fm_Pin18,
    Fm_Pin19,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_PinUndef,
    Fm_Pin32,
    Fm_Pin33,
    Fm_Pin34,
    Fm_Pin35,
    Fm_Pin36,
    Fm_Pin37,
    Fm_Pin38,
    Fm_Pin39
};


/* ポート読み出し許可設定 */
/* 読み出し可能ポートをTRUEに設定する */
const boolean Fm_PortIsReadable[FM_PORT_NUM] = {
    TRUE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
    FALSE,
};


/* BT MAC Address */
boolean BTMacAddr_IsReceived = FALSE;
uint8_t BTMacAddr[6];


/* BT SPP Numeric */
boolean BTSPPNumeric_IsReceived = FALSE;
uint8_t BTSPPNumeric[6];


/**
 * @brief Firmata Reset時処理
 * @note 電源リセット時 or Firmataによるリセットコマンド受信時 or Firmata通信切断時に呼び出される
 */
static void Fm_OnReset(void)
{
    /* 出力を初期化 / 関連アプリケーション初期化処理を追加してください */
    FmLineTrace_Init();
    FmBuzzer_Init();
    FmMotor_Init();
    FmTestApl_Init();

    digitalWrite(PIN_LED1, FALSE);
    digitalWrite(PIN_LED2, FALSE);
    digitalWrite(PIN_LED3, FALSE);
    digitalWrite(PIN_LED4, FALSE);
    digitalWrite(PIN_LED5, FALSE);
    digitalWrite(PIN_LED6, FALSE);
    analogWrite(PIN_MOTOR_L, 0);
    analogWrite(PIN_MOTOR_R, 0);
    analogWrite(PIN_BUZZER, 0);
    analogWrite(PIN_FLED_R, 0);
    analogWrite(PIN_FLED_G, 0);
    analogWrite(PIN_FLED_B, 0);
}

extern uint8_t Fm_RxLength;
extern uint8_t Fm_RxBuffer[BUFFER_SIZE];


/**
 * @brief Execute Offline Command
 */
static void Fm_Execute_OfflineCommand(void)
{
    // 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18
    // F0 71 00 00 00 NN 0X XX 0X XX 0X XX 0X XX 0X XX 0X XX F7
    //       ↑ ↑ ↑Fixed
    // NN=01=MAC
    // NN=02=SSP Numeric
    // 0X XX= 7bit encorded value
    uint8_t ui8t_idx;

    // Special String Command is Received ?
    if (Fm_RxLength == 19 &&
        Fm_RxBuffer[0] == FM_MSGTYPE_SYSEX_START &&
        Fm_RxBuffer[1] == FM_SYSEXCMD_STRING_DATA &&
        Fm_RxBuffer[2] == 0x00 &&
        Fm_RxBuffer[3] == 0x00 &&
        Fm_RxBuffer[18] == FM_MSGTYPE_SYSEX_END)
    {
        // MAC Address Receive ?
        if (Fm_RxBuffer[4] == 0x00 && Fm_RxBuffer[5] == 0x01)
        {
            for (ui8t_idx = 0; ui8t_idx < sizeof(BTMacAddr); ui8t_idx++)
            {
                BTMacAddr[ui8t_idx] = Fm_RxBuffer[6 + ui8t_idx * 2 + 1] << 7 |
                                        Fm_RxBuffer[6 + ui8t_idx * 2];
            }
            BTMacAddr_IsReceived = TRUE;
        }
        // BT SSP Numeric Receive ?
        else if (Fm_RxBuffer[4] == 0x00 && Fm_RxBuffer[5] == 0x02)
        {
            for (ui8t_idx = 0; ui8t_idx < sizeof(BTSPPNumeric); ui8t_idx++)
            {
                BTSPPNumeric[ui8t_idx] = Fm_RxBuffer[6 + ui8t_idx * 2 + 1] << 7 |
                                        Fm_RxBuffer[6 + ui8t_idx * 2];
            }
            BTSPPNumeric_IsReceived = TRUE;
        }
        // Other
        else
        {
        }
    }
}


/**
 * @brief Firmata 初期化処理
 * @note Firmata初期化関数呼び出し時に呼び出される
 */
static void Fm_OnInit(void)
{
    FmLCD_Init();
}


/**
 * @brief Firmata 定期処理
 * @note 10ms間隔で呼び出しされる
 */
static void Fm_OnPeriodic(void)
{
    /* 関連アプリケーション定期処理を追加してください */
    
    FmLineTrace_10msTask();
    FmBuzzer_10msTask();
    FmMotor_10msTask();
    FmTestApl_Periodic();
    FmLCD_10msTask();
}


/**
 * @brief Firmata 接続時処理
 * @note Firmata接続時に1度だけ呼び出しされる
 */
static void Fm_OnConnect(void)
{
    /* SSP Numericを未受信とする */
    BTSPPNumeric_IsReceived = FALSE;
    FmLCD_OnConnect();
}


/**
 * @brief Firmata 切断時処理
 * @note Firmata切断時に1度だけ呼び出しされる
 */
static void Fm_OnDisconnect(void)
{
    FmLCD_OnDisconnect();
}


/*------------------------------------------------------------------------------------------------*/
/* End of Firmata C                                                                               */
/*------------------------------------------------------------------------------------------------*/
#endif


#endif
