/**
 * @file FmLCD.h
 * @brief Scratch用LCD表示処理
 */

#include <board.h>
#include <Std_Types.h>
#include "Firmata.h"
#include "FmLCD.h"


static void FmLCD_DrawOpening(void);
static void FmLCD_DrawConnected(void);
static void FmLCD_DrawDisconnected(void);
static void FmLCD_DrawMonitor(void);


static boolean FmLCD_UpdateReq = FALSE;    /**< LCD Update Req */
static uint32_t FmLCD_UpdateCycle = 0;     /**< モニタ更新タイマ Resolution 10[ms] */
static uint32_t FmLCD_DispTimer = 0;       /**< 画面遷移タイマ Resolution 50[ms] */


#define FmLCD_Disp_OPENING              (0)
#define FmLCD_Disp_DISCONNECTED         (1)
#define FmLCD_Disp_CONNECTED            (2)
#define FmLCD_Disp_APL                  (3)
#define FmLCD_Disp_MONITOR              (4)

#define FmLCD_UPDATE_TIME               (50 / 10)       /**< LCD更新タイマ */
#define FmLCD_Disp_OPENING_TIME         (1000 / 50)     /**< オープニング表示時間 */


static uint32_t FmLCD_Disp;


/**
 * @brief Init
 */
void FmLCD_Init(void)
{
    FmLCD_UpdateCycle = 0;
    FmLCD_DispTimer = 0;
    FmLCD_Disp = FmLCD_Disp_OPENING;
}


/**
 * @brief 10ms定期処理
 */
void FmLCD_10msTask(void)
{
    FmLCD_UpdateCycle++;
    if (FmLCD_UpdateCycle > FmLCD_UPDATE_TIME)
    {
        FmLCD_UpdateCycle = 0;
        
        switch (FmLCD_Disp)
        {
        case FmLCD_Disp_OPENING:
            FmLCD_DrawOpening();
            break;
        case FmLCD_Disp_DISCONNECTED:
            FmLCD_DrawDisconnected();
            break;
        case FmLCD_Disp_CONNECTED:
            FmLCD_DrawConnected();
            break;
        case FmLCD_Disp_APL:
            break;
        case FmLCD_Disp_MONITOR:
            FmLCD_DrawMonitor();
            break;
        default:
            break;
        }
    }


    /* LCD更新要求が残存していたら */
    if (FmLCD_UpdateReq == TRUE)
    {
        /* LCD更新要求を行う */
        FmLCD_UpdateReq = !LcdDrv_update();
    }
}


/**
 * @brief オープニング表示
 */
static void FmLCD_DrawOpening(void)
{
    FmLCD_DispTimer++;
    if (FmLCD_DispTimer > FmLCD_Disp_OPENING_TIME)
    {
        FmLCD_DispTimer = 0;
        FmLCD_Disp = FmLCD_Disp_DISCONNECTED;
    }

    LcdDrv_clear();
    LcdDrv_setCursor(0, 0);
    LcdDrv_print("Scratch iCar");
    LcdDrv_setCursor(1, 0);
    LcdDrv_print("Ver");
    LcdDrv_print(FM_VERSION);

    FmLCD_UpdateReq = !LcdDrv_update();
}


/**
 * @brief 切断時LCD表示処理
 */
static void FmLCD_DrawDisconnected(void)
{
    LcdDrv_clear();
    LcdDrv_setCursor(0, 0);
    LcdDrv_print("Scratch Con.NG");

    if (BTMacAddr_IsReceived == TRUE)
    {
        LcdDrv_setCursor(1, 0);
        LcdDrv_setHex(BTMacAddr[3], 2);
        LcdDrv_setHex(BTMacAddr[4], 2);
        LcdDrv_setHex(BTMacAddr[5], 2);
    }

    if (BTSPPNumeric_IsReceived == TRUE)
    {
        LcdDrv_setCursor(1, 10);
        LcdDrv_setNum(BTSPPNumeric[0], 1);
        LcdDrv_setNum(BTSPPNumeric[1], 1);
        LcdDrv_setNum(BTSPPNumeric[2], 1);
        LcdDrv_setNum(BTSPPNumeric[3], 1);
        LcdDrv_setNum(BTSPPNumeric[4], 1);
        LcdDrv_setNum(BTSPPNumeric[5], 1);
    }

    FmLCD_UpdateReq = !LcdDrv_update();
}


/**
 * @brief 接続時LCD表示処理
 */
static void FmLCD_DrawConnected(void)
{
    LcdDrv_clear();
    LcdDrv_setCursor(0, 0);
    LcdDrv_print("Scratch Con.OK!");

    FmLCD_UpdateReq = !LcdDrv_update();
}


/**
 * @brief センサモニタ表示
 */
static void FmLCD_DrawMonitor(void)
{
    LcdDrv_setCursor(0, 0);
    LcdDrv_print("L");
    LcdDrv_setNum(analogRead(PIN_LINE_L), 4);

    LcdDrv_print(" ");
    LcdDrv_setNum(analogRead(PIN_DISTANCE), 4);

    LcdDrv_print(" R");
    LcdDrv_setNum(analogRead(PIN_LINE_R), 4);

    LcdDrv_setCursor(1, 0);
    LcdDrv_print("V");
    LcdDrv_setNum(analogRead(PIN_VOLUME), 3);

    LcdDrv_print(" B");
    LcdDrv_setNum(analogRead(PIN_BRIGHTNESS), 4);

    LcdDrv_print(" SW");
    LcdDrv_setNum(analogRead(PIN_TOGGLE), 1);
    LcdDrv_setNum(digitalRead(PIN_BTN1), 1);
    LcdDrv_setNum(digitalRead(PIN_BTN2), 1);

    FmLCD_UpdateReq = !LcdDrv_update();
}


/**
 * @brief モニタ有効化
 * @param isEnable 有効=1, 無効=0
 */
void FmLCD_SetMonitorEnable(int isEnable)
{
    if (isEnable)
    {
        FmLCD_Disp = FmLCD_Disp_MONITOR;
    }
    else
    {
        FmLCD_Disp = FmLCD_Disp_APL;
    }
}


/**
 * @brief ユーザテキストを設定
 * @param text \0終端文字列データへのポインタ
 */
void FmLCD_SetUserText(const char *lcdData)
{
    FmLCD_Disp = FmLCD_Disp_APL;
    
    /* Disable LCDMonitor */
    FmLCD_SetMonitorEnable(0);
    
    LcdDrv_setCursor(0, 0);
    LcdDrv_print(lcdData);
    FmLCD_UpdateReq = !LcdDrv_update();
}


/**
 * @brief 接続通知
 */
void FmLCD_OnConnect(void)
{
    FmLCD_Disp = FmLCD_Disp_CONNECTED;
}


/**
 * @brief 切断通知
 */
void FmLCD_OnDisconnect(void)
{
    FmLCD_Disp = FmLCD_Disp_DISCONNECTED;
}
